import struct
from . import consts
from . import errors


__all__ = ['encode_command', 'Reader']


class Reader(object):
    """This class is responsible for parsing replies from the stream
    of data that is read from a *Gibson* connection. It does not contain
    functionality to handle I/O"""

    def __init__(self):
        self._buffer = bytearray()
        self._payload = bytearray()
        self._is_header = False
        self._is_payload = False
        self._resp_size = None
        self._gb_encoding = None
        self._code = None

    def feed_data(self, data):
        """Put raw chunk of data obtained from connection to buffer.

        :param data: ``bytes``, raw input data.
        """
        if not data:
            return
        self._buffer.extend(data)

    def gets(self):
        """When the buffer does not contain a full reply, gets returns
        False. This means extra data is needed and feed should be called
        again before calling gets again:

        :return: ``False`` there is no full reply or parsed obj.
        """
        if not self._is_header and len(self._buffer) >= consts.HEADER_SIZE:
            reply = self._buffer[:consts.HEADER_SIZE]
            unpacked = struct.unpack(b'<HBI', reply)
            self._code, self._gb_encoding, self._resp_size = unpacked
            self._is_header = True

        if (self._is_header and not self._is_payload and
                (len(self._buffer) >= self._resp_size + consts.HEADER_SIZE)):

            start = consts.HEADER_SIZE
            end = consts.HEADER_SIZE + self._resp_size
            data = self._buffer[start:end]
            self._payload.extend(data)
            self._is_payload = True

        if self._is_header and self._is_payload:
            values = self._parse_replay()
            self._reset()
            return values
        return False

    def _parse_replay(self):
        if self._code == consts.REPL_ERR:
            return errors.GibsonServerError()
        elif self._code == consts.REPL_OK:
            return True
        elif self._code == consts.REPL_ERR_NOT_FOUND:
            return None
        elif self._code == consts.REPL_ERR_NAN:
            return errors.ExpectedANumber()
        elif self._code == consts.REPL_ERR_MEM:
            return errors.MemoryLimitError()
        elif self._code == consts.REPL_ERR_LOCKED:
            return errors.KeyLockedError()
        elif self._code == consts.REPL_VAL:
            return self._parse_value(self._payload, self._gb_encoding)
        elif self._code == consts.REPL_KVAL:
            return self._parse_kv(self._payload)
        else:
            raise errors.ProtocolError()

    def _parse_kv(self, data):
        # parse key/value replay from Gibson server
        pairs_num = struct.unpack('I', data[:consts.REPL_SIZE])[0]
        result = []
        offset = consts.REPL_SIZE

        for i in range(pairs_num):
            # unpack key size
            key_size = \
                struct.unpack('I', data[offset: offset + consts.REPL_SIZE])[0]
            offset += consts.REPL_SIZE
            # unpack key
            key = bytes(data[offset: offset + key_size])
            result.append(key)
            offset += key_size
            # unpack value encoding
            value_gb_encodig = struct.unpack('B', data[offset: offset + 1])[0]
            offset += 1
            # unpack value size
            value_size = \
                struct.unpack('I', data[offset: offset + consts.REPL_SIZE])[0]
            offset += consts.REPL_SIZE
            # unpack value
            value = self._parse_value(data[offset: offset + value_size],
                                      value_gb_encodig)
            result.append(value)
            offset += value_size
        return result

    def _parse_value(self, data, encoding):
        # parse simple value replay from Gibson server.
        # apply gibson encoding if needed
        if encoding == consts.GB_ENC_NUMBER:
            return struct.unpack('q', data)[0]
        elif encoding == consts.GB_ENC_PLAIN:
            return bytes(data)
        else:
            raise errors.ProtocolError()

    def _reset(self):
        # replay parsed, buffer should be prepared for next server
        # reply
        self._buffer = self._buffer[consts.HEADER_SIZE + self._resp_size:]
        self._payload = bytearray()
        self._is_header = False
        self._is_payload = False
        self._resp_size = None
        self._gb_encoding = None
        self._code = None


_converters = {
    bytes: lambda val: val,
    bytearray: lambda val: val,
    str: lambda val: val.encode('utf-8'),
    int: lambda val: str(val).encode('utf-8'),
    float: lambda val: str(val).encode('utf-8'),
    }


def encode_command(command, *args):
    """Pack and encode *gibson* command acording to gibson binary protocol

    :see: http://gibson-db.in/protocol/

    :param command: ``bytes``, gibson command (get, set, etc.)
    :param args: required arguments for given command.
    :return: ``bytes`` packed and encoded command.
    """
    _args = []
    for arg in args:
        if type(arg) in _converters:
            _args.append(_converters[type(arg)](arg))
        else:
            raise TypeError("Argument {!r} expected to be of bytes,"
                            " str, int or float type".format(arg))

    op_code = consts.command_map[command]
    query = b' '.join(_args)
    fmt = '<IH{}s'.format(len(query))
    data = struct.pack(fmt, consts.OP_CODE_SIZE + len(query), op_code, query)
    return data



from .connection import GibsonConnection, create_connection
from .errors import (GibsonError, ProtocolError, ReplyError,
                     ExpectedANumber, MemoryLimitError, KeyLockedError)
from .pool import GibsonPool, create_pool, create_gibson

__version__ = '0.1.1'

# make pyflakes happy
(GibsonConnection, create_connection, GibsonError, ProtocolError, ReplyError,
    ExpectedANumber, MemoryLimitError, KeyLockedError, GibsonPool,
    create_pool, create_gibson)



__all__ = [
    'GibsonError',
    'ProtocolError',
    'ReplyError',
    'ExpectedANumber',
    'MemoryLimitError',
    'KeyLockedError',
    ]


class GibsonError(Exception):
    """Base exception class for aiogibson exceptions.
    """


class ProtocolError(GibsonError):
    """Raised when protocol error occurs.
    """


class ReplyError(GibsonError):
    """Generic error while executing the query"""


class GibsonServerError(GibsonError):
    """Unknown error on gibson server"""


class ExpectedANumber(GibsonError):
    """Expected a number ( TTL or TIME ) but the specified value
    was invalid."""


class MemoryLimitError(GibsonError):
    """The server reached configuration memory limit and will not accept
    any new value until its freeing routine will be executed."""


class KeyLockedError(GibsonError):
    """The specified key was locked by a `OP_LOCK` or a `OP_MLOCK` query."""



import asyncio

from .connection import create_connection

__all__ = ['create_gibson', 'Gibson']


class Gibson:
    """High-level Gibson interface

    :see: http://gibson-db.in/commands/
    """

    def __init__(self, connection):
        self._conn = connection

    def __repr__(self):
        return '<Gibson {!r}>'.format(self._conn)

    def close(self):
        self._conn.close()

    @property
    def closed(self):
        """True if connection is closed."""
        return self._conn.closed

    @asyncio.coroutine
    def get(self, key):
        """Get the value for a given key.

        :param key: ``bytes``  key to get.
        :return: ``bytes`` if value exists else ``None``
        """
        return (yield from self._conn.execute(b'get', key))

    @asyncio.coroutine
    def set(self, key, value, expire=0):
        """Set the value for the given key, with an optional TTL.

        :param key: ``bytes`` key to set.
        :param value: ``bytes`` value to set.
        :param expire: ``int``  optional ttl in seconds
        :raises TypeError: if expire argument is not ``int``
        """
        if not isinstance(expire, int):
            raise TypeError('expire must be int')
        return (yield from self._conn.execute(b'set', expire, key, value))

    @asyncio.coroutine
    def delete(self, key):
        """ Delete the given key.

        :param key: ``bytes`` key to delete.
        :return: ``bool`` true in case of success.
        """
        result = (yield from self._conn.execute(b'del', key))
        return bool(result)

    @asyncio.coroutine
    def ttl(self, key, expire):
        """Set the TTL of a key.

        :param key: ``bytes``, key to set ttl.
        :param expire: ``int``, TTL in seconds.
        :return: ``bool``, True in case of success.
        :raises TypeError: if expire argument is not ``int``
        """
        if not isinstance(expire, int):
            raise TypeError('expire must be int')
        result = yield from self._conn.execute(b'ttl', key, expire)
        return bool(result)

    @asyncio.coroutine
    def inc(self, key):
        """Increment by one the given key.

        :param key: ``bytes``, key to increment.
        :return: ``int`` incremented value
        """
        return (yield from self._conn.execute(b'inc', key))

    @asyncio.coroutine
    def dec(self, key):
        """Decrement by one the given key.

        :param key: ``bytes``, key to decrement.
        :return: ``int`` decremented value in case of success
        """
        return (yield from self._conn.execute(b'dec', key))

    @asyncio.coroutine
    def lock(self, key, expire=0):
        """Prevent the given key from being modified for a given amount
        of seconds.

        :param key: ``bytes``, key to decrement.
        :param expire: ``int``, time in seconds to lock the item.
        :return: ``bool``
        :raises TypeError: if expire argument is not ``int``
        """
        if not isinstance(expire, int):
            raise TypeError('expire must be int')
        result = yield from self._conn.execute(b'lock', key, expire)
        return bool(result)

    @asyncio.coroutine
    def unlock(self, key):
        """Remove the lock from the given key.

        :param key: ``bytes`` key ot unlock.
        :return: ``bool``, True in case of success.
        """
        result = yield from self._conn.execute(b'unlock', key)
        return bool(result)

    @asyncio.coroutine
    def keys(self, prefix):
        """Return a list of keys matching the given prefix.

        :param prefix: key prefix to use as expression.
        :return: ``list`` of available keys
        """
        result = yield from self._conn.execute(b'keys', prefix)
        keys = [r for i, r in enumerate(result) if i % 2]
        return keys

    @asyncio.coroutine
    def stats(self):
        """Get system stats about the Gibson instance.

        :return: ``list`` of pairs (stat, value).
        """
        result = yield from self._conn.execute(b'stats')
        return result

    @asyncio.coroutine
    def ping(self):
        """Ping the server instance to refresh client last seen timestamp.

        :return: ``True`` or error.
        """
        return (yield from self._conn.execute(b'ping'))

    @asyncio.coroutine
    def meta_size(self, key):
        """The size in bytes of the item value.

        :param key: ``bytes``, key of interest.
        :return: ``int``, value size in bytes
        """
        return (yield from self._conn.execute(b'meta', key, b'size'))

    @asyncio.coroutine
    def meta_encoding(self, key):
        """Gibson encoding for given value.

        :param key: ``bytes``, key of interest.
        :return: ``int``, gibson encoding, 0 - ``bytes``, 2 - ``int``.
        """
        return (yield from self._conn.execute(b'meta', key, b'encoding'))

    @asyncio.coroutine
    def meta_access(self, key):
        """Timestamp of the last time the item was accessed.

        :param key: ``bytes``, key of interest.
        :return: ``int``, timestamp
        """
        return (yield from self._conn.execute(b'meta', key, b'access'))

    @asyncio.coroutine
    def meta_created(self, key):
        """Timestamp of item creation.

        :param key: ``bytes``, key of interest.
        :return: ``int``, timestamp
        """
        return (yield from self._conn.execute(b'meta', key, b'created'))

    @asyncio.coroutine
    def meta_ttl(self, key):
        """Item specified time to live, -1 for infinite TTL.

        :param key: ``bytes``, key of interest.
        :return: ``int``, seconds of TTL.
        """
        return (yield from self._conn.execute(b'meta', key,  b'ttl'))

    @asyncio.coroutine
    def meta_left(self, key):
        """Number of seconds left for the item to live if a ttl
        was specified, otherwise -1.

        :param key: ``bytes``, key of interest.
        :return: ``int``, Number of seconds left.
        """
        return (yield from self._conn.execute(b'meta', key, b'left'))

    @asyncio.coroutine
    def meta_lock(self, key):
        """Number of seconds the item is locked, -1 if there's no lock.

        :param key: ``bytes``, key of interest.
        :return: ``int``, number of seconds
        """
        return (yield from self._conn.execute(b'meta', key, b'lock'))

    @asyncio.coroutine
    def end(self):
        """Disconnects from the client from gibson instance."""
        return (yield from self._conn.execute(b'end'))

    @asyncio.coroutine
    def mset(self, prefix, value):
        """Set the value for keys verifying the given prefix.

        :param prefix: prefix for keys.
        :return: ``int``, number of modified items, otherwise an error.
        """
        return (yield from self._conn.execute(b'mset', prefix, value))

    @asyncio.coroutine
    def mget(self, prefix):
        """Get the values for keys with given prefix.

        :param prefix: prefix for keys.
        :return: ``list`` of key/value pairs
        """
        return (yield from self._conn.execute(b'mget', prefix))

    @asyncio.coroutine
    def mttl(self, prefix, expire=0):
        """Set the TTL for keys verifying the given prefix.

        :param prefix: prefix for keys.
        :param expire: ``int``, new expiration time.
        :return: ``int``, number of modified items, otherwise an error.
        :raises TypeError: if expire argument is not ``int``
        """
        if not isinstance(expire, int):
            raise TypeError('expire must be int')
        return (yield from self._conn.execute(b'mttl', prefix, expire))

    @asyncio.coroutine
    def minc(self, prefix):
        """Increment by one keys verifying the given prefix.

        :param prefix: prefix for keys.
        :return: ``int``, number of modified items, otherwise an error.
        """
        return (yield from self._conn.execute(b'minc', prefix))

    @asyncio.coroutine
    def mdec(self, prefix):
        """Decrement by one keys verifying the given prefix.

        :param prefix: prefix for keys.
        :return: ``int``, number of modified items, otherwise an error.
        """
        return (yield from self._conn.execute(b'mdec', prefix))

    @asyncio.coroutine
    def mlock(self, prefix, expire=0):
        """Prevent keys verifying the given prefix from being modified
        for a given amount of seconds.

        :param prefix: ``bytes``, prefix for keys.
        :param expire:``int``, lock period in seconds.
        :return: ``int``, number of modified items, otherwise an error.
        :raises TypeError: if expire argument is not ``int``
        """
        if not isinstance(expire, int):
            raise TypeError('expire must be int')
        result = yield from self._conn.execute(b'mlock', prefix, expire)
        return result

    @asyncio.coroutine
    def munlock(self, prefix):
        """Remove the lock on keys verifying the given prefix.

        :param prefix: prefix for keys.
        :return: ``int``, number of affected items, otherwise an error.
        """
        result = yield from self._conn.execute(b'munlock', prefix)
        return result

    @asyncio.coroutine
    def mdelete(self, prefix):
        """Delete keys verifying the given prefix.

        :param prefix: prefix for keys.
        :return: ``int``, number of modified items, otherwise an error.
        """
        result = yield from self._conn.execute(b'mdel', prefix)
        return result

    @asyncio.coroutine
    def count(self, prefix):
        """Count items for a given prefix.

        :param prefix: ``bytes`` The key prefix to use as expression
        :return: ``int`` number of elements
        """
        return (yield from self._conn.execute(b'count', prefix))


@asyncio.coroutine
def create_gibson(address, *, encoding=None, commands_factory=Gibson,
                  loop=None):
    """Create high-level Gibson interface.

    :param address: ``str`` for unix socket path, or ``tuple``
        for (host, port) tcp connection.
    :param encoding: this argument can be used to decode byte-replies to
        strings. By default no decoding is done.
    :param commands_factory:
    :param loop: event loop to use
    :return: high-level Gibson connection ``Gibson``
    """
    conn = yield from create_connection(address, encoding=encoding, loop=loop)
    return commands_factory(conn)



"""
Borrowed from aioredi.

:see: https://github.com/aio-libs/aioredis/blob/master/aioredis/pool.py
"""

import asyncio

from .commands import create_gibson, Gibson

__all__ = ['create_pool', 'GibsonPool']


@asyncio.coroutine
def create_pool(address, *, encoding=None, minsize=10, maxsize=10,
                commands_factory=Gibson, loop=None):
    """Creates Gibson Pool.

    By default it creates pool of commands_factory instances, but it is
    also possible to create pool of plain connections by passing
    ``lambda conn: conn`` as commands_factory.
    All arguments are the same as for create_connection.
    Returns GibsonPool instance.
    """

    pool = GibsonPool(address, encoding=encoding,
                      minsize=minsize, maxsize=maxsize,
                      commands_factory=commands_factory,
                      loop=loop)
    yield from pool._fill_free()
    return pool


class GibsonPool:
    """Gibson connections pool.
    """

    def __init__(self, address, encoding=None,
                 *, minsize, maxsize, commands_factory, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._address = address
        self._minsize = minsize
        self._factory = commands_factory
        self._loop = loop
        self._pool = asyncio.Queue(maxsize, loop=loop)
        self._used = set()
        self._encoding = encoding

    @property
    def minsize(self):
        """Minimum pool size.
        """
        return self._minsize

    @property
    def maxsize(self):
        """Maximum pool size.
        """
        return self._pool.maxsize

    @property
    def size(self):
        """Current pool size.
        """
        return self.freesize + len(self._used)

    @property
    def freesize(self):
        """Current number of free connections.
        """
        return self._pool.qsize()

    @asyncio.coroutine
    def clear(self):
        """Clear pool connections.

        Close and remove all free connections.
        """
        while not self._pool.empty():
            conn = yield from self._pool.get()
            conn.close()

    @property
    def encoding(self):
        """Current set codec or None."""
        return self._encoding

    @asyncio.coroutine
    def acquire(self):
        """Acquires a connection from free pool.

        Creates new connection if needed.
        """
        yield from self._fill_free()
        if self.minsize > 0 or not self._pool.empty():
            conn = yield from self._pool.get()
        else:
            conn = yield from self._create_new_connection()
        assert not conn.closed, conn
        assert conn not in self._used, (conn, self._used)
        self._used.add(conn)
        return conn

    def release(self, conn):
        """Returns used connection back into pool.
        """
        assert conn in self._used, "Invalid connection, maybe from other pool"
        self._used.remove(conn)
        if not conn.closed:
            try:
                self._pool.put_nowait(conn)
            except asyncio.QueueFull:
                # consider this connection as old and close it.
                conn.close()

    @asyncio.coroutine
    def _fill_free(self):
        while self.freesize < self.minsize and self.size < self.maxsize:
            conn = yield from self._create_new_connection()
            yield from self._pool.put(conn)

    @asyncio.coroutine
    def _create_new_connection(self):
        conn = yield from create_gibson(self._address,
                                        encoding=self._encoding,
                                        commands_factory=self._factory,
                                        loop=self._loop)
        return conn

    def __enter__(self):
        raise RuntimeError(
            "'yield from' should be used as a context manager expression")

    def __exit__(self, *args):
        pass    # pragma: nocover

    def __iter__(self):
        # this method is needed to allow `yield`ing from pool
        conn = yield from self.acquire()
        return _ConnectionContextManager(self, conn)


class _ConnectionContextManager:

    __slots__ = ('_pool', '_conn')

    def __init__(self, pool, conn):
        self._pool = pool
        self._conn = conn

    def __enter__(self):
        return self._conn

    def __exit__(self, exc_type, exc_value, tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None



"""
Constants for gibson binary protocol

:see: http://gibson-db.in/protocol/
:see: https://github.com/evilsocket/gibson/blob/unstable/src/query.h
"""

# OPERATION CODES
# single
OP_SET = 0x01
OP_TTL = 0x02
OP_GET = 0x03
OP_DEL = 0x04
OP_INC = 0x05
OP_DEC = 0x06
OP_LOCK = 0x07
OP_UNLOCK = 0x08
# multi
OP_MSET = 0x09
OP_MTTL = 0x0A
OP_MGET = 0x0B
OP_MDEL = 0x0C
OP_MINC = 0x0D
OP_MDEC = 0x0E
OP_MLOCK = 0x0F
OP_MUNLOCK = 0x10
# other
OP_COUNT = 0x11
OP_STATS = 0x12
OP_PING = 0x13
OP_META = 0x14
OP_KEYS = 0x15
OP_END = 0xFF


command_map = {
    b'set': OP_SET,
    b'ttl': OP_TTL,
    b'get': OP_GET,
    b'del': OP_DEL,
    b'inc': OP_INC,
    b'dec': OP_DEC,
    b'lock': OP_LOCK,
    b'unlock': OP_UNLOCK,

    b'mset': OP_MSET,
    b'mttl': OP_MTTL,
    b'mget': OP_MGET,
    b'mdel': OP_MDEL,
    b'minc': OP_MINC,
    b'mdec': OP_MDEC,
    b'mlock': OP_MLOCK,
    b'munlock': OP_MUNLOCK,
    b'count': OP_COUNT,

    b'stats': OP_STATS,
    b'ping': OP_PING,
    b'meta': OP_META,
    b'keys': OP_KEYS,
    b'end': OP_END
}


# ENCODING
# Raw string data follows.
GB_ENC_PLAIN = 0x00
# Compressed data, this is a **reserved** value not used for replies.
GB_ENC_LZF = 0x01
# Packed **long** number follows, four bytes for 32bit architectures,
# eight bytes for 64bit.
GB_ENC_NUMBER = 0x02


# **Reply**
REPL_ERR = 0x00
REPL_ERR_NOT_FOUND = 0x01
REPL_ERR_NAN = 0x02
REPL_ERR_MEM = 0x03
REPL_ERR_LOCKED = 0x04
REPL_OK = 0x05
REPL_VAL = 0x06
REPL_KVAL = 0x07


OP_CODE_SIZE = 2
ENCODING_SIZE = 1
REPL_SIZE = 4  # size of block where size of data stored
HEADER_SIZE = OP_CODE_SIZE + ENCODING_SIZE + REPL_SIZE



"""
Borrowed from aioredis.

:see: https://github.com/aio-libs/aioredis/blob/master/aioredis/connection.py
"""

import asyncio

from .errors import GibsonError, ProtocolError
from .parser import Reader, encode_command


__all__ = ['create_connection', 'GibsonConnection']

MAX_CHUNK_SIZE = 65536
_NOTSET = object()


@asyncio.coroutine
def create_connection(address, *, encoding=None, loop=None):
    """Creates GibsonConnection connection.
    Opens connection to Gibson server specified by address argument.

    :param address: ``str`` for unix socket path, or ``tuple``
        for (host, port) tcp connection.
    :param encoding: this argument can be used to decode byte-replies to
        strings. By default no decoding is done.
    """
    assert isinstance(address, (tuple, list, str)), "tuple or str expected"

    if isinstance(address, (list, tuple)):
        host, port = address
        reader, writer = yield from asyncio.open_connection(
            host, port, loop=loop)
    else:
        reader, writer = yield from asyncio.open_unix_connection(
            address, loop=loop)
    conn = GibsonConnection(reader, writer, address=address,
                            encoding=encoding, loop=loop)
    return conn


class GibsonConnection:
    """Gibson connection."""

    def __init__(self, reader, writer, address, *, encoding=None, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._reader = reader
        self._writer = writer
        self._loop = loop
        self._waiters = asyncio.Queue(loop=self._loop)
        self._parser = Reader()
        self._reader_task = asyncio.Task(self._read_data(), loop=self._loop)
        self._closing = False
        self._closed = False
        self._address = address
        self._encoding = encoding

    def __repr__(self):
        return '<GibsonConnection {}>'.format(self._address)

    @asyncio.coroutine
    def _read_data(self):
        """Responses reader task."""
        while not self._reader.at_eof():
            data = yield from self._reader.read(MAX_CHUNK_SIZE)
            self._parser.feed_data(data)
            while True:
                try:
                    obj = self._parser.gets()
                except ProtocolError as exc:
                    # ProtocolError is fatal
                    # so connection must be closed
                    self._closing = True
                    self._loop.call_soon(self._do_close, exc)
                    return
                else:
                    if obj is False:
                        break
                    waiter = yield from self._waiters.get()
                    if isinstance(obj, GibsonError):
                        waiter.set_exception(obj)
                    else:
                        waiter.set_result(obj)
        self._closing = True
        self._loop.call_soon(self._do_close, None)

    @asyncio.coroutine
    def execute(self, command, *args, encoding=_NOTSET):
        """Executes raw gibson command.

        :param command: ``str`` or ``bytes`` gibson command.
        :param args: tuple of arguments required for gibson command.
        :param encoding: ``str`` default encoding for unpacked data.

        :raises TypeError: if any of args can not be encoded as bytes.
        :raises ProtocolError: when response can not be decoded meaning
            connection is broken.
        """

        assert self._reader and not self._reader.at_eof(), (
            "Connection closed or corrupted")

        if command is None:
            raise TypeError("command must not be None")
        if None in set(args):
            raise TypeError("args must not contain None")

        command = command.strip()
        data = encode_command(command, *args)
        self._writer.write(data)
        yield from self._writer.drain()
        fut = asyncio.Future(loop=self._loop)
        yield from self._waiters.put(fut)
        result = yield from fut
        if encoding is _NOTSET:
            encoding = self._encoding
        if encoding is not None and isinstance(result, bytes):
            return result.decode(encoding)
        return result

    def close(self):
        """Close connection."""
        self._do_close(None)

    def _do_close(self, exc):
        if self._closed:
            return
        self._closed = True
        self._closing = False
        self._writer.transport.close()
        self._reader_task.cancel()
        self._reader_task = None
        self._writer = None
        self._reader = None
        while self._waiters.qsize():
            waiter = self._waiters.get_nowait()
            if exc is None:
                waiter.cancel()
            else:
                waiter.set_exception(exc)

    @property
    def closed(self):
        """True if connection is closed."""
        closed = self._closing or self._closed
        if not closed and self._reader and self._reader.at_eof():
            self._closing = closed = True
            self._loop.call_soon(self._do_close, None)
        return closed

    @property
    def encoding(self):
        """Current set codec or None."""
        return self._encoding
